# apache-lib.pl
# Common functions for apache configuration

  do '/usr/webtool/web-lib.pl';
  &init_config();

  %access = &get_module_acl();
  @access_types = $access{'types'} eq '*' ? (0 .. 100)
                                          : split(/\s+/, $access{'types'});
  map { $access_types{$_}++ } @access_types;

  ### Read the site-specific setup file, then require in all the module-specific
  ### .pl files
  if (&read_file("$module_config_directory/site", \%site)) {
    local($m, $f, $d);
    $httpd_size = $site{'size'};
    foreach $m (split(/\s+/, $site{'modules'})) {
      if ($m =~ /(\S+)\/(\S+)/) { $httpd_modules{$1} = $2; }
    }
    foreach $f (split(/\s+/, $site{'htaccess'})) {
      if (-r $f) { push(@htaccess_files, $f); }
    }
    foreach $m (keys %httpd_modules) {
      do "./$m.pl";
    }
    foreach $d (split(/\s+/, $site{'defines'})) {
      $httpd_defines{$d}++;
    }
  }

# parse_config_file(handle, lines, file)
# Parses lines of text from some config file into a data structure. The
# return value is an array of references, one for each directive in the file.
# Each reference points to an associative array containing
#  line -	The line number this directive is at
#  eline -	The line number this directive ends at
#  file -	The file this directive is from
#  type -	0 for a normal directive, 1 for a container directive
#  name -	The name of this directive
#  value -	Value (possibly with spaces)
#  members -	For type 1, a reference to the array of members
sub parse_config_file {
  local($fh, @rv, $line, %dummy);
  $fh = $_[0];

  $dummy{'line'} = $dummy{'eline'} = $_[1]-1;
  $dummy{'file'} = $_[2];
  $dummy{'type'} = 0;
  $dummy{'name'} = "dummy";

  @rv = (\%dummy);
  while($line = <$fh>) {
    chop;
    $line =~ s/^\s*#.*$//g;
    if ($line =~ /^\s*<\/(\S+)\s*(.*)>/) {
      ### end of a container directive. This can only happen in a
      ### recursive call to this function
      $_[1]++;
      last;
    }
    elsif ($line =~ /^\s*<IfModule\s+(\!?)(\S+)\.c>/i) {
      ### start of an IfModule block. Read it, and if the module
      ### exists put the directives in this section.
      local ($not, $mod) = ($1, $2);
      local $oldline = $_[1];
      $_[1]++;
      local @dirs = &parse_config_file($fh, $_[1], $_[2]);

      if (!$not && $httpd_modules{$mod} || $not && !$httpd_modules{$mod}) {
        ### use the directives..
        push(@rv, { 'line', $oldline, 'eline', $oldline, 'file', $_[2],
                    'name', "<IfModule $not$mod>" }); push(@rv, @dirs);
        push(@rv, { 'line', $_[1]-1, 'eline', $_[1]-1, 'file', $_[2],
                    'name', "</IfModule>" });
      }
    }
    elsif ($line =~ /^\s*<IfDefine\s+(\!?)(\S+)>/i) {
      ### start of an IfDefine block. Read it, and if the define
      ### exists put the directives in this section
      local ($not, $def) = ($1, $2);
      local $oldline = $_[1];
      $_[1]++;
      local @dirs = &parse_config_file($fh, $_[1], $_[2]);

      if (!$not && $httpd_defines{$def} || $not && !$httpd_defines{$def}) {
        ### use the directives..
        push(@rv, { 'line', $oldline, 'eline', $oldline, 'file', $_[2],
                    'name', "<IfDefine $not$def>" });
        push(@rv, @dirs);
        push(@rv, { 'line', $_[1]-1, 'eline', $_[1]-1, 'file', $_[2],
                    'name', "</IfDefine>" });
      }
    }
    elsif ($line =~ /^\s*<(\S+)\s*(.*)>/) {
      ### start of a container directive. The first member is a dummy
      ### directive at the same line as the container
      local(%dir, @members);
      %dir = ('line', $_[1], 'file', $_[2], 'type', 1, 'name', $1, 'value', $2);
      $dir{'value'} =~ s/\s+$//g;
      $dir{'words'} = &wsplit($dir{'value'});
      $_[1]++;
      @members = &parse_config_file($fh, $_[1], $_[2]);
      $dir{'members'} = \@members;
      $dir{'eline'} = $_[1]-1;
      push(@rv, \%dir);
    }
    elsif ($line =~ /^\s*(\S+)\s*(.*)$/) {
      ### normal directive
      local(%dir);
      %dir = ('line', $_[1], 'eline', $_[1], 'file', $_[2],
              'type', 0, 'name', $1, 'value', $2);

      if ($dir{'value'} =~ s/\\$//g) {
        ### multi-line directive!
        while($line = <$fh>) {
          chop($line);
          $cont = ($line =~ s/\\$//g);
          $dir{'value'} .= $line;
          $dir{'eline'} = ++$_[1];
          if (!$cont) { last; }
        }
      }

      $dir{'value'} =~ s/\s+$//g;
      $dir{'words'} = &wsplit($dir{'value'});
      push(@rv, \%dir);
      $_[1]++;
    }
    else {
      ### blank or comment line
      $_[1]++;
    }
  }

  return @rv;
}

# wsplit(string)
# Splits a string like  foo "foo \"bar\"" bazzz  into an array of words
sub wsplit
{
local($s, @rv); $s = $_[0];
$s =~ s/\\\"/\0/g;
while($s =~ /^"([^"]*)"\s*(.*)$/ || $s =~ /^(\S+)\s*(.*)$/) {
	$w = $1; $s = $2;
	$w =~ s/\0/"/g; push(@rv, $w);
	}
return \@rv;
}

# wjoin(word, word, ...)
sub wjoin
{
local(@rv, $w);
foreach $w (@_) {
	if ($w =~ /^\S+$/) { push(@rv, $w); }
	else { push(@rv, "\"$w\""); }
	}
return join(' ', @rv);
}


### find_directive(name, &directives)
### Returns the values of directives matching some name
sub find_directive {
  local(@rv, $i, @vals, $dref);

  foreach $ref (@{$_[1]}) {
    if (lc($ref->{'name'}) eq lc($_[0])) { push(@vals, $ref->{'value'}); }
  }

  return wantarray ? @vals : !@vals ? undef : $vals[$#vals];
}

# find_directive_struct(name, &directives)
# Returns references to directives maching some name
sub find_directive_struct
{
local(@rv, $i, @vals);
foreach $ref (@{$_[1]}) {
	if (lc($ref->{'name'}) eq lc($_[0])) {
		push(@vals, $ref);
		}
	}
return wantarray ? @vals : !@vals ? undef : $vals[$#vals];
}

# find_vdirective(name, &virtualdirectives, &directives)
# Looks for some directive in a <VirtualHost> section, and then in the 
# main section
sub find_vdirective
{
if ($_[1]) {
	$rv = &find_directive($_[0], $_[1]);
	if ($rv) { return $rv; }
	}
return &find_directive($_[0], $_[2]);
}

# get_config()
# Returns the entire config structure
sub get_config
{
  local($acc, $res, $lnum, $conf, @virt, $v, $mref, $inc);

if (@get_config_cache) { return \@get_config_cache; }

# read primary config file
$conf = $config{'httpd_conf'};
if (!$conf) { $conf = "$config{'httpd_dir'}/conf/httpd.conf"; }
if (!-r $conf) { $conf = "$config{'httpd_dir'}/etc/httpd.conf"; }
open(CONF, $conf) || return undef;
$lnum = 0;
@get_config_cache = &parse_config_file(CONF, $lnum,$conf);
close(CONF);

# Read main resource and access config files
$lnum = 0;
$res = &find_directive("ResourceConfig", \@get_config_cache);
if (!$res) { $res = $config{'srm_conf'}; }
if (!$res) { $res = "$config{'httpd_dir'}/conf/srm.conf"; }
if (!-r $res) { $res = "$config{'httpd_dir'}/etc/srm.conf"; }
open(CONF, $res);
push(@get_config_cache, &parse_config_file(CONF, $lnum, $res));
close(CONF);

$lnum = 0;
$acc = &find_directive("AccessConfig", \@get_config_cache);
if (!$acc) { $acc = $config{'access_conf'}; }
if (!$acc) { $acc = "$config{'httpd_dir'}/conf/access.conf"; }
if (!-r $acc) { $acc = "$config{'httpd_dir'}/etc/access.conf"; }
open(CONF, $acc);
push(@get_config_cache, &parse_config_file(CONF, $lnum, $acc));
close(CONF);

# Read extra config files in VirtualHost sections
@virt = &find_directive_struct("VirtualHost", \@get_config_cache);
foreach $v (@virt) {
	$mref = $v->{'members'};
	if ($res = &find_directive("ResourceConfig", $mref)) {
		if ($res !~ /^\//) { $res = "$config{'httpd_dir'}/$res"; }
		$lnum = 0;
		open(CONF, $res);
		push(@$mref, &parse_config_file(CONF, $lnum, $res));
		close(CONF);
		}
	if ($acc = &find_directive("AccessConfig", $mref)) {
		if ($acc !~ /^\//) { $acc = "$config{'httpd_dir'}/$acc"; }
		$lnum = 0;
		open(CONF, $acc);
		push(@$mref, &parse_config_file(CONF, $lnum, $acc));
		close(CONF);
		}
	}

# Read Include'd files
foreach $inc (&find_directive("Include", \@get_config_cache)) {
	if ($inc !~ /^\//) { $inc = "$config{'httpd_dir'}/$inc"; }
	$lnum = 0;
	open(CONF, $inc);
	push(@get_config_cache, &parse_config_file(CONF, $lnum, $inc));
	close(CONF);
	}

# Move directives from active IfModule sections to section they are in

return \@get_config_cache;
}

### get_virtual_config(index)
sub get_virtual_config {
  local($conf, $c, $v);
  $conf = &get_config();
  if (!$_[0]) { $c = $conf; $v = undef; }
  else {
    $c = $conf->[$_[0]]->{'members'};
    $v = $conf->[$_[0]];
  }

  return wantarray ? ($c, $v) : $c;
}

# get_htaccess_config(file)
sub get_htaccess_config {
  local($lnum, @conf);

  open(HTACCESS, $_[0]);
    @conf = &parse_config_file(HTACCESS, $lnum, $_[0]);
  close(HTACCESS);

  return \@conf;
}

### ############################################################################
### save_directive(name, &values, &directives, &config)
### Updates the config file(s) and the directives structure with new values
### for the given directives.
### If a directive's value is merely being changed, then its value only needs
### to be updated in the directives array and in the file.
### ############################################################################
sub save_directive {
  local($i, @old, $lref, $change, $len);

  @old = &find_directive_struct($_[0], $_[2]);

  for($i=0; $i<@old || $i<@{$_[1]}; $i++) {
	$v = ${$_[1]}[$i];
	if ($i >= @old) {
		# a new directive is being added. If other directives of this
		# type exist, add it after them. Otherwise, put it at the end of
		# the first file in the section
		if ($change) {
			# Have changed some old directive.. add this new one
			# after it, and update change
			local(%v, $j);
			%v = (	"line", $change->{'line'}+1,
				"eline", $change->{'line'}+1,
				"file", $change->{'file'},
				"type", 0,
				"name", $_[0],
				"value", $v);
			$j = &indexof($change, @{$_[2]})+1;
			&renumber($_[3], $v{'line'}, $v{'file'}, 1);
			splice(@{$_[2]}, $j, 0, \%v);
			$lref = &read_file_lines($v{'file'});
			splice(@$lref, $v{'line'}, 0, "$_[0] $v");
			$change = \%v;
			}
		else {
			# Adding a new directive to the end of the list
			# in this section
			local($f, %v, $j);
			$f = $_[2]->[0]->{'file'};
			for($j=0; $_[2]->[$j]->{'file'} eq $f; $j++) { }
			$l = $_[2]->[$j-1]->{'eline'}+1;
			%v = (	"line", $l,
				"eline", $l,
				"file", $f,
				"type", 0,
				"name", $_[0],
				"value", $v);
			&renumber($_[3], $l, $f, 1);
			splice(@{$_[2]}, $j, 0, \%v);
			$lref = &read_file_lines($f);
			splice(@$lref, $l, 0, "$_[0] $v");
			}
		}
	elsif ($i >= @{$_[1]}) {
		# a directive was deleted
		$lref = &read_file_lines($old[$i]->{'file'});
		$idx = &indexof($old[$i], @{$_[2]});
		splice(@{$_[2]}, $idx, 1);
		$len = $old[$i]->{'eline'} - $old[$i]->{'line'} + 1;
		splice(@$lref, $old[$i]->{'line'}, $len);
		&renumber($_[3], $old[$i]->{'line'}, $old[$i]->{'file'}, -$len);
		}
	else {
		# just changing the value
		$lref = &read_file_lines($old[$i]->{'file'});
		$len = $old[$i]->{'eline'} - $old[$i]->{'line'} + 1;
		&renumber($_[3], $old[$i]->{'eline'}+1,
			  $old[$i]->{'file'},1-$len);
		$old[$i]->{'value'} = $v;
		$old[$i]->{'eline'} = $old[$i]->{'line'};
		splice(@$lref, $old[$i]->{'line'}, $len, "$_[0] $v");
		$change = $old[$i];
		}
	}
}

# renumber(&config, line, file, offset)
# Recursively changes the line number of all directives from some file 
# beyond the given line.
sub renumber
{
local($d);
if (!$_[3]) { return; }
foreach $d (@{$_[0]}) {
	if ($d->{'file'} eq $_[2] && $d->{'line'} >= $_[1]) {
		$d->{'line'} += $_[3];
		}
	if ($d->{'file'} eq $_[2] && $d->{'eline'} >= $_[1]) {
		$d->{'eline'} += $_[3];
		}
	if ($d->{'type'}) {
		&renumber($d->{'members'}, $_[1], $_[2], $_[3]);
		}
	}
}

# server_root(path, &directives)
sub server_root
{
if ($_[0] =~ /^\//) { return $_[0]; }
else { return "$config{'httpd_dir'}/$_[0]"; }
}

sub dump_config
{
local($c, $mref);
print "<table border>\n";
print "<tr> <td>Name</td> <td>Value</td> <td>File</td> <td>Line</td> </tr>\n";
foreach $c (@_) {
	printf "<tr> <td>%s</td> <td>%s</td><td>%s</td><td>%s</td> </tr>\n",
		$c->{'name'}, $c->{'value'}, $c->{'file'}, $c->{'line'};
	if ($c->{'type'}) {
		print "<tr> <td colspan=4>\n";
		$mref = $c->{'members'};
		&dump_config(@$mref);
		print "</td> </tr>\n";
		}
	}
print "</table>\n";
}

sub def
{
return $_[0] ? $_[0] : $_[1];
}

# make_directives(ref, version, module)
sub make_directives
{
local(@rv, $aref);
$aref = $_[0];
local $ver = $_[1];
if ($ver =~ /^(1)\.(3)(\d+)$/) {
	$ver = sprintf "%d.%d%2.2d", $1, $2, $3;
	}
foreach $d (@$aref) {
	local(%dir);
	$dir{'name'} = $d->[0];
	$dir{'multiple'} = $d->[1];
	$dir{'type'} = $d->[2];
	$dir{'module'} = $_[2];
	$dir{'version'} = $ver;
	$dir{'priority'} = $d->[5];
	foreach $c (split(/\s+/, $d->[3])) { $dir{$c}++; }
	if (!$d->[4]) { push(@rv, \%dir); }
	elsif ($d->[4] =~ /^-([\d\.]+)$/ && $ver < $1) { push(@rv, \%dir); }
	elsif ($d->[4] =~ /^([\d\.]+)$/ && $ver >= $1) { push(@rv, \%dir); }
	elsif ($d->[4] =~ /^([\d\.]+)-([\d\.]+)$/ && $ver >= $1 && $ver < $2)
		{ push(@rv, \%dir); }
	}
return @rv;
}


# editable_directives(type, context)
# Returns an array of references to associative arrays, one for each 
# directive of the given type that can be used in the given context
sub editable_directives
{
local($m, $func, @rv);
foreach $m (keys %httpd_modules) {
	$func = $m."_directives";
	push(@rv, &$func($httpd_modules{$m}));
	}
@rv = grep { $_->{'type'} == $_[0] && $_->{$_[1]} } @rv;
@rv = sort { $pd = $b->{'priority'} - $a->{'priority'};
	     $md = $a->{'module'} cmp $b->{'module'};
	     $pd == 0 ? ($md == 0 ? $a->{'name'} cmp $b->{'name'} : $md) : $pd }
		@rv;
return @rv;
}


### ############################################################################
### generate_inputs(&editors, &directives)
### Displays a 2-column list of options, for use inside a table
### ############################################################################
sub generate_inputs {
  local($e, $sw, @args, @rv, $func);

  foreach $e (@{$_[0]}) {
    if (!$sw) { print "<tr>\n"; }

    ### Build arg list for the editing function. Each arg can be a single
    ### directive struct, or a reference to an array of structures.
    $func = "edit";
    undef(@args);

    foreach $ed (split(/\s+/, $e->{'name'})) {
      local(@vals);
      $func .= "_$ed";
      @vals = &find_directive_struct($ed, $_[1]);
      if ($e->{'multiple'}) { push(@args, \@vals); }
      elsif (!@vals) { push(@args, undef); }
      else { push(@args, $vals[$#vals]); }
    }

    push(@args, $e);

    ### call the function
    @rv = &$func(@args);

    if ($rv[0] == 2) {
      ### spans 2 columns..
      ### need to end this row
      if ($sw) { print "<td colspan=2></td> </tr><tr>\n"; }
      else { $sw = !$sw; }

      print "  <td valign=top align=left><DIV CLASS=nbody><b>$rv[1]</b></td>\n";
      print "</TR>\n\n";

      print "<TR>\n";
      print "  <td nowrap valign=top colspan=3 width=80%><DIV CLASS=tbody>$rv[2]</td>\n";
      print "</TR>\n\n";

      print "<TR>\n";
      print "  <TD COLSPAN=4>&nbsp;</TD>\n";
    }
    else {
      ### only spans one column
      print "  <td valign=top><DIV CLASS=nbody><b>$rv[1]</b></td>\n";
      print "</TR>\n\n";

      print "<TR>\n";
      print "  <td nowrap valign=top width=25%><DIV CLASS=tbody>$rv[2]</td>\n";
      print "</TR>\n\n";

      print "<TR>\n";
      print "  <TD COLSPAN=4>&nbsp;</TD>\n";
      print "</TR>\n\n";

      print "<TR>\n";
    }

    if ($sw) { print "</tr>\n"; }
    $sw = !$sw;
  }
}



### parse_inputs(&editors, &directives, &config)
### Reads user choices from a form and update the directives and config files.
sub parse_inputs {
  ### First call editor functions to get new values. Each function returns an
  ### array of references to arrays containing the new values for the directive.
  foreach $e (@{$_[0]}) {
    @dirs = split(/\s+/, $e->{'name'});
    $func = "save_".join('_', @dirs);
    @rv = &$func($e);

    for($i=0; $i<@dirs; $i++) {
      push(@chname, $dirs[$i]);
      push(@chval, $rv[$i]);
    }
  }

  ### Assuming everything went OK, update the configuration
  for($i=0; $i<@chname; $i++) {
    &save_directive($chname[$i], $chval[$i], $_[1], $_[2]);
  }

  &flush_file_lines();
}


### opt_input(value, name, default, size)
sub opt_input {
return sprintf "<FONT FACE=Arial,Helvetica,Sans-Serif SIZE=2>\n".
               "<input type=radio name=$_[1]_def value=1 %s> $_[2]\n".
	       "<input type=radio name=$_[1]_def value=0 %s></FONT>\n".
	       "<input name=$_[1] size=$_[3] value='%s'>\n",
	defined($_[0]) ? "" : "checked",
	defined($_[0]) ? "checked" : "",
	$_[0];
}

# parse_opt(name, regexp, error)
sub parse_opt
{
local($i, $re);
if ($in{"$_[0]_def"}) { return ( [ ] ); }
for($i=1; $i<@_; $i+=2) {
	$re = $_[$i];
	if ($in{$_[0]} !~ /$re/) { &error($_[$i+1]); }
	}
return ( [ $in{$_[0]} ] );
}

# choice_input(value, name, default, [choice]+)
# Each choice is a display,value pair
sub choice_input {
  local($i, $rv);

  for($i=3; $i<@_; $i++) {
    $_[$i] =~ /^([^,]*),(.*)$/;
    $rv .= "<FONT FACE=Arial,Helvetica,Sans-Serif SIZE=2>";
    $rv .= sprintf "<input type=radio name=$_[1] value=\"$2\" %s> $1\n",
           lc($2) eq lc($_[0]) ||
              !defined($_[0]) && lc($2) eq lc($_[2]) ? "checked" : "";
    $rv .= "<BR>\n";
  }

  return $rv;
}

# choice_input_vert(value, name, default, [choice]+)
# Each choice is a display,value pair
sub choice_input_vert
{
local($i, $rv);
for($i=3; $i<@_; $i++) {
	$_[$i] =~ /^([^,]*),(.*)$/;
	$rv .= sprintf "<input type=radio name=$_[1] value=\"$2\" %s> $1<br>\n",
		lc($2) eq lc($_[0]) || !defined($_[0]) &&
				       lc($2) eq lc($_[2]) ? "checked" : "";
	}
return $rv;
}

# parse_choice(name, default)
sub parse_choice
{
if (lc($in{$_[0]}) eq lc($_[1])) { return ( [ ] ); }
else { return ( [ $in{$_[0]} ] ); }
}

# select_input(value, name, default, [choice]+)
sub select_input {
  local($i, $rv);

  $rv = "<DIV CLASS=nbody><select name=\"$_[1]\">\n";
for($i=3; $i<@_; $i++) {
	$_[$i] =~ /^([^,]*),(.*)$/;
	$rv .= sprintf "<option value=\"$2\" %s> $1\n",
		lc($2) eq lc($_[0]) || !defined($_[0]) && lc($2) eq lc($_[2]) ? "selected" : "";
	}
$rv .= "</select>\n";
return $rv;
}

# parse_choice(name, default)
sub parse_select
{
return &parse_choice(@_);
}

# handler_input(value, name)
sub handler_input
{
local($m, $func, @hl, $rv, $h);
push(@hl, "");
foreach $m (keys %httpd_modules) {
	$func = $m."_handlers";
	if (defined(&$func)) {
		push(@hl, &$func($conf, $httpd_modules{$m}));
		}
	}
if (&indexof($_[0], @hl) < 0) { push(@hl, $_[0]); }
$rv = "<select name=$_[1]>\n";
foreach $h (&unique(@hl)) {
	$rv .= sprintf "<option value=\"$h\" %s>$h\n",
		$h eq $_[0] ? "selected" : "";
	}
$rv .= "</select>\n";
return $rv;
}

### parse_handler(name)
sub parse_handler {
  if ($in{$_[0]} eq "") { return ( [ ] ); }
  else { return ( [ $in{$_[0]} ] ); }
}

### virtual_name(struct)
sub virtual_name {
  if ($_[0]) {
    local $n = &find_directive("ServerName", $_[0]->{'members'});
    return $n ? $n : $_[0]->{'value'};
  }
  else { return "default server"; }
}

### dir_name(struct)
sub dir_name {
  local($dfm, $mat);
  $_[0]->{'name'} =~ /^(Directory|Files|Location)(Match)?$/;
  $dfm = $1; $mat = $2;

  if ($mat) { return "$dfm regexp <tt>$_[0]->{'words'}->[0]</tt>"; }
  elsif ($_[0]->{'words'}->[0] eq "~") {
    return "$dfm regexp <tt>$_[0]->{'words'}->[1]</tt>";
  }
  else { return "$dfm <tt>$_[0]->{'words'}->[0]</tt>"; }
}

### list_user_file(file, &user,  &pass)
sub list_user_file {
  local($_);

  open(USERS, $_[0]);
    while(<USERS>) { /^(\S+):(\S+)/; push(@{$_[1]}, $1); $_[2]->{$1} = $2; }
  close(USERS);
}



### ############################################################################
### config_icons(context, program)
### Displays up to 17 icons, one for each type of configuration directive, for
### some context (global, virtual, directory or htaccess)
### <RWM>
### ############################################################################
sub config_icons {
  local($m, $func, $e, %etype, $i, $c);
  foreach $m (sort { $a cmp $b } (keys %httpd_modules)) {
    $func = $m."_directives";
      foreach $e (&$func($httpd_modules{$m})) {
        if ($e->{$_[0]}) { $etype{$e->{'type'}}++; }
      }
  }

  print "<CENTER><TABLE BORDER=0 WIDTH=620 CELLPADDING=0 CELLSPACING=0>\n";

  for($i=0; $text{"type_$i"}; $i++) {
    if ($etype{$i} && $access_types{$i}) {
      print "<TR>\n";
      print "  <TD><DIV CLASS=nbody><B>";
      print "<A HREF=\"$_[1]type=$i\">", $text{"type_$i"}, "</A></TD>\n";
      print "  <TD><DIV CLASS=nbody>", $text{"helptext_$i"}, "</TD>\n";
      print "</TR>\n";
      print "<TR><TD>&nbsp;</TD></TR>\n";
      $c++;
    }
  }

  for($i=2; $i<@_; $i++) {
    print "<TR>\n";
    print "  <TD><DIV CLASS=nbody><B>\n";
    print "<A HREF=\"", $_[$i]->{'link'}, "\">", $_[$i]->{'name'},"</A></TD>\n";
    #print "  <TD><DIV CLASS=nbody>", $text{"helptext_$i"}, "</TD>\n";
    print "  <TD><DIV CLASS=nbody>$_[$i]->{'desc'}</TD>\n";
    print "</TR>\n";
    print "<TR><TD>&nbsp;</TD></TR>\n";
    $c++;
  }
  print "</TABLE></CENTER><p>\n";
}





# restart_button()
# Returns HTML for a link to put in the top-right corner of every page
sub restart_button
{
local($pidfile, $pidfilestr, $conf, $rv);
$conf = &get_config();
$pidfilestr = &find_directive_struct("PidFile", $conf);
$pidfile = $pidfilestr ? $pidfilestr->{'words'}->[0]
		       : "logs/httpd.pid";
$pidfile = &server_root($pidfile, $conf);
$args = "redir=".&urlize(&this_url());
if (open(PID, $pidfile) && <PID> =~ /(\d+)/ && kill(0, $1)) {
	close(PID);
	$rv = "<a href=\"restart.cgi?$args\">$text{'apache_apply'}</a><br>\n";
	if ($access{'stop'}) {
		$rv .= "<a href=\"stop.cgi?$args\">$text{'apache_stop'}</a>\n";
		}
	}
elsif ($access{'stop'}) {
	$rv = "<a href=\"start.cgi?$args\">$text{'apache_start'}</a><br>\n";
	}
return $rv;
}

# this_url()
# Returns the URL in the apache directory of the current script
sub this_url
{
local($url);
$url = $ENV{'SCRIPT_NAME'};
if ($ENV{'QUERY_STRING'} ne "") { $url .= "?$ENV{'QUERY_STRING'}"; }
return $url;
}

# find_all_directives(config, name)
# Recursively finds all directives of some type
sub find_all_directives
{
local(@rv, $d);
foreach $d (@{$_[0]}) {
	if ($d->{'name'} eq $_[1]) { push(@rv, $d); }
	if ($d->{'type'} == 1) {
		push(@rv, &find_all_directives($d->{'members'}, $_[1]));
		}
	}
return @rv;
}

# httpd_info(executable)
# Returns the httpd version and modules array
sub httpd_info
{
local(@mods, $verstr, $ver, $minor);
$verstr = `$_[0] -v 2>&1`;
if ($verstr =~ /Apache(\S*)\/(\d+)\.([\d\.]+)/) {
	# standard apache
	$ver = $2; $minor = $3; $minor =~ s/\.//g; $ver .= ".$minor";
	}
elsif ($verstr =~ /Red\s*Hat\s+Secure\/2\.0/i) {
	# redhat secure server 2.0
	$ver = 1.31;
	}
elsif ($verstr =~ /Red\s*Hat\s+Secure\/3\.0/i) {
	# redhat secure server 3.0
	$ver = 1.39;
	}
else {
	# couldn't get version
	return (0, undef);
	}
if ($ver < 1.2) {
	# apache 1.1 has no -l option! Use the standard list
	@mods = ("core", "mod_mime", "mod_access", "mod_auth", "mod_include",
		 "mod_negotiation", "mod_dir", "mod_cgi", "mod_userdir",
		 "mod_alias", "mod_env", "mod_log_common");
	}
else {
	# ask apache for the module list
	@mods = ("core");
	open(APACHE, "$_[0] -l |");
	while(<APACHE>) {
		if (/(mod_\S+)\.c/) { push(@mods, $1); }
		}
	close(APACHE);
	}
return ($ver, \@mods);
}

# print_line(directive, text, indent, link)
sub print_line
{
local $line = $_[0]->{'line'} + 1;
local $lstr = "$_[0]->{'file'} ($line)";
local $txt = join("", @{$_[1]});
local $left = 85 - length($lstr) - $_[2];
if (length($txt) > $left) {
	$txt = substr($txt, 0, $left)." ..";
	}
local $txtlen = length($txt);
$txt =~ s/&/&amp;/g;
$txt =~ s/</&lt;/g;
$txt =~ s/>/&gt;/g;
print " " x $_[2];
if ($_[3]) {
	print "<a href=\"$_[3]\">",$txt,"</a>";
	}
else { print $txt; }
print " " x (90 - $txtlen - $_[2] - length($lstr));
print $lstr,"\n";
}

### can_edit_virt(struct)
sub can_edit_virt {
  return 1 if ($access{'virts'} eq '*');
  local %vcan;
  map { $vcan{$_}++ } split(/\s+/, $access{'virts'});
  return $vcan{&virt_acl_name($_[0])};
}

### virt_acl_name(struct)
sub virt_acl_name {
  return "__default__" if (!$_[0]);
  local $n = &find_directive("ServerName", $_[0]->{'members'});
  if ($n) {
    local $p;
    if ($_[0]->{'value'} =~ /(:\d+)/) { $p = $1; }
    return $n.$p;
  }
  else { return $_[0]->{'value'}; }
}

### allowed_auth_file(file)
sub allowed_auth_file {
  local $_;
  return 1 if ($access{'dir'} eq '/');
  return 0 if ($_[0] =~ /\.\./);
  local $f = &server_root($_[0], &get_config());
  return 0 if (-l $f && !&allowed_auth_file(readlink($f)));
  local $l = length($access{'dir'});
  return length($f) >= $l && substr($f, 0, $l) eq $access{'dir'};
}



sub Generate_Certificate {
  local($form);
  $form = @_ > 1 ? $_[1] : 0;

  return "<input type=button onClick='ifield = document.forms[$form].$_[0]; chooser = window.open(\"/VHost/cert_gen.cgi?file=\"+ifield.value, \"chooser\", \"toolbar=no,menubar=no,scrollbar=no,width=720,height=700\"); chooser.ifield = ifield' value=\"Generate Certificate\">\n";
}


sub get_nvhs() {

  &obtain_lock("/etc/httpd/conf/httpd.conf");
  open CONF, "/etc/httpd/conf/httpd.conf";

  while (<CONF>) {
    if ($_ =~ m/^NameVirtualHost\ (.*)$/) { push @candidates, $1; }
  }

  close CONF;
  &release_lock("/etc/httpd/conf/httpd.conf");

  return @candidates;
}


  ### ###############################################################
  ### PrintError()
  ### ###############################################################
  ### Used to pop up a JavaScript with a $message then return user to
  ### previous page.
  ### ###############################################################

  sub PrintError {
    my ($message) = @_;

    print <<ErrorEOF;
Content-type: text/html


<HTML>
<HEAD>
  <TITLE> </TITLE>

  <SCRIPT LANGUAGE="JavaScript">
    alert("$message");
    history.go(-1);
  </SCRIPT>
</HEAD>
</HTML>
ErrorEOF

  exit;
  }

# XXX:  This should probably be ripped out at some point.
  ### ##########################################################################
  sub return_squirrelmail_config()
  {
    local ($org_name, $domain, $imapServerAddress, $smtpServerAddress, $rv);
    ($org_name, $domain, $imapServerAddress, $smtpServerAddress)  =  @_;

    $rv  = <<EOF;
<?php
\t/** SquirrelMail configuration
\t/** Created by the Guardian Digital WebTool.
\t ** Based upon the configure script, conf.pl
\t **/

if (defined('config_php'))
   return;
define('config_php', true);

\tglobal \$print_config_version, \$config_version, \$config_use_color;
\t\$config_version = "x63";
\t\$config_use_color = 1;

\tglobal \$org_name, \$org_logo, \$org_title, \$signout_page;
\t\$org_name   = "$org_name";
\t\$org_logo   = "../images/logo.png";
\t\$org_title  = "SquirrelMail \$version";
\t\$signout_page  = "";

\tglobal \$domain, \$imapServerAddress, \$imapPort;
\tglobal \$useSendmail, \$smtpServerAddress, \$smtpPort;
\tglobal \$sendmail_path, \$use_authenticated_smtp, \$imap_server_type;
\tglobal \$invert_time, \$optional_delimiter;
\t\$domain                 = "$domain";
\t\$imapServerAddress      = "$imapServerAddress";
\t\$imapPort               =  143;
\t\$useSendmail            =  false;
\t\$smtpServerAddress      = "$smtpServerAddress";
\t\$smtpPort               =  25;
\t\$sendmail_path          = "/usr/sbin/sendmail";
\t\$use_authenticated_smtp = false;
\t\$imap_server_type       = "uw";
\t\$invert_time            = false;
\t\$optional_delimiter     = "detect";

\tglobal \$default_folder_prefix, \$trash_folder, \$sent_folder;
\tglobal \$default_move_to_trash, \$default_move_to_sent;
\tglobal \$show_prefix_option, \$list_special_folders_first;
\tglobal \$use_special_folder_color, \$auto_expunge, \$default_sub_of_inbox;
\tglobal \$show_contain_subfolders_option, \$default_unseen_notify;
\tglobal \$default_unseen_type, \$auto_create_special;
\t\$default_folder_prefix            = "";
\t\$trash_folder                     = "INBOX.Trash";
\t\$sent_folder                      = "INBOX.Sent";
\t\$default_move_to_trash            =  true;
\t\$default_move_to_sent             =  true;
\t\$show_prefix_option               =  false;
\t\$list_special_folders_first       =  true;
\t\$use_special_folder_color         =  true;
\t\$auto_expunge                     =  true;
\t\$default_sub_of_inbox             =  true;
\t\$show_contain_subfolders_option   =  false;
\t\$default_unseen_notify            =  2;
\t\$default_unseen_type              =  1;
\t\$auto_create_special              =  true;

\tglobal \$default_charset, \$data_dir, \$attachment_dir;
\tglobal \$default_left_size, \$force_username_lowercase;
\t\$default_charset          = "iso-8859-1";
\t\$data_dir                 = "../data/";
\t\$attachment_dir           = "\$data_dir";
\t\$default_left_size        =  150;
\t\$force_username_lowercase = false;

\tglobal \$plugins;

\tglobal \$theme_css, \$theme;
\t\$theme_css = "../themes/css/squirrel.css";
\t\$theme[0]["PATH"] = "../themes/default_theme.php";
\t\$theme[0]["NAME"] = "Default";
\t\$theme[1]["PATH"] = "../themes/plain_blue_theme.php";
\t\$theme[1]["NAME"] = "Plain Blue";
\t\$theme[2]["PATH"] = "../themes/sandstorm_theme.php";
\t\$theme[2]["NAME"] = "Sand Storm";
\t\$theme[3]["PATH"] = "../themes/deepocean_theme.php";
\t\$theme[3]["NAME"] = "Deep Ocean";
\t\$theme[4]["PATH"] = "../themes/slashdot_theme.php";
\t\$theme[4]["NAME"] = "Slashdot";
\t\$theme[5]["PATH"] = "../themes/purple_theme.php";
\t\$theme[5]["NAME"] = "Purple";
\t\$theme[6]["PATH"] = "../themes/forest_theme.php";
\t\$theme[6]["NAME"] = "Forest";
\t\$theme[7]["PATH"] = "../themes/ice_theme.php";
\t\$theme[7]["NAME"] = "Ice";
\t\$theme[8]["PATH"] = "../themes/seaspray_theme.php";
\t\$theme[8]["NAME"] = "Sea Spray";
\t\$theme[9]["PATH"] = "../themes/bluesteel_theme.php";
\t\$theme[9]["NAME"] = "Blue Steel";
\t\$theme[10]["PATH"] = "../themes/dark_grey_theme.php";
\t\$theme[10]["NAME"] = "Dark Grey";
\t\$theme[11]["PATH"] = "../themes/high_contrast_theme.php";
\t\$theme[11]["NAME"] = "High Contrast";
\t\$theme[12]["PATH"] = "../themes/black_bean_burrito_theme.php";
\t\$theme[12]["NAME"] = "Black Bean Burrito";
\t\$theme[13]["PATH"] = "../themes/servery_theme.php";
\t\$theme[13]["NAME"] = "Servery";
\t\$theme[14]["PATH"] = "../themes/maize_theme.php";
\t\$theme[14]["NAME"] = "Maize";
\t\$theme[15]["PATH"] = "../themes/bluesnews_theme.php";
\t\$theme[15]["NAME"] = "BluesNews";
\t\$theme[16]["PATH"] = "../themes/deepocean2_theme.php";
\t\$theme[16]["NAME"] = "Deep Ocean 2";

\tglobal \$default_use_javascript_addr_book, \$ldap_server;
\t\$default_use_javascript_addr_book = false;
\tglobal \$motd;
\t\$motd = "";
\tglobal \$squirrelmail_default_language;
\t\$squirrelmail_default_language = "en";
EOF

    return $rv;
  }
1;
