#!/usr/bin/perl

=begin comment info
 +-----------------------------------------------------------------------------+
 | Copyright 2001, Guardian Digital Inc.
 | $Id: create_virt.cgi,v 1.3.4.2 2003/04/18 12:37:09 rwm Exp $
 |
 | Ryan W. Maple, <ryan@guardiandigital.com>
 +-----------------------------------------------------------------------------+
=end comment info
=cut

  ###  Import our library routines.
  require './apache-lib.pl';
  &ReadParse();

  $conf = &get_config();
  @virt = &find_directive_struct("VirtualHost", $conf);
  if (@virt >= 10) {
    my $msg = "You have reached the limit of 10 virtual hosts available\\n"
            . "in the Community Edition of EnGarde Secure Linux.";
    &PrintError($msg);
    exit;
  }

  ### ##########################################################################
  ### BEGIN input validation.
  ###

    ###  Setup IP Address.
    if (&check_ipaddress($in{'addr'})) { $addr = $in{'addr'}; }
    else { &PrintError("'$in{'addr'}' is not a valid IP address."); exit; }

    ###  Setup Port.
    if (!$in{'port'}) { $port = ":80";  $in{'port'} = 80; }
    else              { $port = ":443"; $in{'port'} = 443; }

    ###   Setup ServerAdmin.
    if ($in{'admin'} =~ m/^.*@.*$/) { $admin = $in{'admin'}; }
    else { &PrintError("'$in{'admin'}' is not a valid admin e-mail."); exit; }

    ###   Setup ServerName.
    if ($in{'name'} =~ m/^[A-Za-z0-9\-\.]+$/) { $name = $in{'name'}; }
    else { &PrintError("'$in{'name'}' is not a valid hostname."); exit; }

    ###   Setup Webmaster.
    if (getpwnam($in{'webmaster'})) { $webmaster = $in{'webmaster'}; }
    else { &PrintError("'$in{'webmaster'}' is not a valid user."); exit; }

    ###   Setup Group.
    if (getgrnam($in{'group'})) { $group = $in{'group'}; }
    else { &PrintError("'$in{'group'}' is not a valid group."); exit; }

    ###   Setup "base directory".  Throw an error if it exists.
    $root = "/home/httpd/$name-$in{'port'}";
    if (-d $root) { &PrintError("$root already exists."); exit; }

    ###  If we want to enable webmail, make sure everything else is valid.
    if ($in{'webmail_enable'})
    {

      ###  Make sure we have an "Organization Name"
      if (!$in{'webmail_organization'}) {
        &PrintError("Please enter a Webmail 'Organization Name'"); exit;
      }

      ###  Make sure we have a "Domain Name"
      elsif (!$in{'webmail_domain'}) {
        &PrintError("Please enter a Webmail 'Domain Name'"); exit;
      }

      ###  Make sure our "Domain Name" is valid.
      elsif ($in{'webmail_domain'} !~ m/^[A-Za-z0-9\-\.]+$/) {
        &PrintError("You entered an invalid Webmail 'Domain Name'"); exit;
      }

      ###  Make sure we have an "IMAP Server"
      elsif (!$in{'webmail_imap'}) {
        &PrintError("Please enter a Webmail 'IMAP Server'"); exit;
      }

      ###  Make sure our "IMAP Server" is valid.
      elsif ($in{'webmail_imap'} !~ m/^[A-Za-z0-9\-\.]+$/) {
        &PrintError("You entered an invalid Webmail 'IMAP SMTP'"); exit;
      }

      ###  Make sure we have an "SMTP Server"
      elsif (!$in{'webmail_smtp'}) {
        &PrintError("Please enter a Webmail 'SMTP Server'"); exit;
      }

      ###  Make sure our "SMTP Server" is valid.
      elsif ($in{'webmail_smtp'} !~ m/^[A-Za-z0-9\-\.]+$/) {
        &PrintError("You entered an invalid Webmail 'SMTP Server'"); exit;
      }

    }


  ### ##########################################################################
  ### Database Setup ("normal" virtual hosts ONLY).
  ### 
  if ($in{'create_db'}) {
 
    if ($in{'port'} == 443) {
      &PrintError("You can not currently create a database for an SSL site.");
      exit;
    }
 
    ### Check integrity of username. 
    if ($in{'db_uname'} =~ m/^[A-Za-z0-9]{1,16}$/) {
      $db_uname = $in{'db_uname'};
    }
    else {
      &PrintError("'$in{'db_uname'}' is not a valid database username.");
      exit;
    }

    ### Check integrity of password. 
    if ($in{'db_pass'} =~ m/^[^\'\"\/]{1,16}$/) {
      $db_pass = $in{'db_pass'};
    }
    else {
     $in{'db_pass'} =~ s/\"/\\\"/g;
     &PrintError("'$in{'db_pass'}' is not a valid database password.");
     exit;
    }


    ### Make the database name.
    $db_name = $in{'name'}; $db_name =~ s/\./\_/g;

    ### Let's use the DBI since it's SO much more fun...
    use DBI;

    ### Get root's mysql password.
    open FILE, "/etc/webtool/SystemManager/config";
      while (<FILE>) { if ($_ =~ m/^pass\=(.*)$/) { $mysql_pass = $1; } }
    close FILE;

    ### Setup DBMS variables.
    $database = "mysql"; $host = "localhost"; $user = "root";


    ### Connect to the database.
    my $dbh = DBI->connect("DBI:mysql:database=$database;host=$host", "$user",
                           "$mysql_pass");


    ### Create the database.
    $sth = $dbh->prepare("CREATE DATABASE $db_name");
    $sth->execute();
    $sth->finish();


    ### Create the user.
    $sth = $dbh->prepare("INSERT INTO mysql.user VALUES ('localhost',
                         '$db_uname', PASSWORD('$db_pass'), 'N', 'N', 'N', 'N',
                         'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N', 'N')");
    $sth->execute();
    $sth->finish();


    ### Create the database permissions.
    $sth = $dbh->prepare("INSERT INTO mysql.db VALUES ('localhost', '$db_name',
                         '$db_uname', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y',
                         'Y', 'Y')");
    $sth->execute();
    $sth->finish();


    ### Reload the grant tables.
    $sth = $dbh->prepare("FLUSH PRIVILEGES");
    $sth->execute();
    $sth->finish();


    ### Disconnect from the DBMS.
    $dbh->disconnect();
  }


  ### ##########################################################################
  ### httpd.conf Setup
  ###
  $f = $config{'virt_file'};
  if (!$f) {
    $vconf = &get_virtual_config();
    $f = $vconf->[0]->{'file'};
    for($j=0; $vconf->[$j]->{'file'} eq $f; $j++) { }
    $l = $vconf->[$j-1]->{'eline'}+1;
  }

  $lref = &read_file_lines($f);
    push(@$lref, "");
    push(@$lref, "<VirtualHost $addr$port>");
    push(@$lref, "<Directory \"$root/html\">");
    push(@$lref, "  AllowOverride all");
    push(@$lref, "</Directory>");
    push(@$lref, "");
    push(@$lref, "<Directory \"$root/cgi-bin\">");
    push(@$lref, "  Options +ExecCGI");
    push(@$lref, "</Directory>");
    push(@$lref, "");

    push(@$lref, "DirectoryIndex index.php") if ($in{'webmail_enable'});

    push(@$lref, "DocumentRoot $root/html");
    push(@$lref, "ServerName $name");
    push(@$lref, "ServerAdmin $admin");
    push(@$lref, "ScriptAlias /cgi-bin/ $root/cgi-bin/");
    push(@$lref, "ErrorLog $root/logs/error_log");
    push(@$lref, "TransferLog $root/logs/access_log");

    if ($port eq ":443") {
      push(@$lref, "\nSSLEngine off");
      push(@$lref, "SSLLog $root/logs/ssl_engine_log");
    }

    push(@$lref, "</VirtualHost>");
    push(@$lref, "");
  &flush_file_lines();


  ### ##########################################################################
  ### Directory Structure Setup
  ###
  mkdir "$root",         0755;
  mkdir "$root/cgi-bin", 0755;
  mkdir "$root/html",    0755;
  mkdir "$root/logs",    0755;

  ###  Only create/chmod the /ssl/ directory if this is in fact an SSL host.
  mkdir "$root/ssl",     0755   if ($port eq ":443");



  ### ##########################################################################
  ### Webstats Setup ("normal" virtual hosts ONLY)
  ###
  if ($port ne ":443") {
    ### Setup the WEBSTATS directory.
    system("/bin/mkdir --mode=0775 -p $root/html/WEBSTATS > /dev/null 2>&1");
    system "/bin/chmod 2751 $root/html/WEBSTATS > /dev/null 2>&1;";

    ### Create a Webalizer configuration file.
    if (!-r "/etc/webalizer") { mkdir "/etc/webalizer", 0755; }

    open WEBSTATS, ">/etc/webalizer/$name.conf";
      print WEBSTATS "LogFile\t$root/logs/access_log\n";
      print WEBSTATS "OutputDir\t$root/html/WEBSTATS\n";
      print WEBSTATS "HostName\t$name\n\n";
  
      print WEBSTATS "LogType\tweb\n";
    close WEBSTATS;

    ### Create a Webalizer .htacess and .htpasswd.
    system "/bin/touch $root/html/WEBSTATS/.htpasswd";
    open HTACCESS, ">$root/html/WEBSTATS/.htaccess";
      print HTACCESS "AuthUserFile $root/html/WEBSTATS/.htpasswd\n";
      print HTACCESS "AuthName \"Web Stats for $name\"\n";
      print HTACCESS "AuthType Basic\n\n";

      print HTACCESS "<Limit GET>\n";
      print HTACCESS "require valid-user\n";
      print HTACCESS "</Limit>\n\n";
    close HTACCESS;
  }


  ### ##########################################################################
  ### Webmail Setup (SSL virtual hosts ONLY)
  ###
  if ($in{'webmail_enable'})
  {
    ###  Create and populate the /webmail/ directory.
    mkdir "$root/html/webmail", 0755;
    system "chmod 0755 $root/html/webmail                     > /dev/null 2>&1";
    system "cp -a /usr/lib/squirrelmail/* $root/html/webmail/ > /dev/null 2>&1";

    ###  Write our configuration.
    open CONFIG, ">$root/html/webmail/config/config.php";
      print CONFIG &return_squirrelmail_config($in{'webmail_organization'},
                                               $in{'webmail_domain'},
                                               $in{'webmail_imap'},
                                               $in{'webmail_smtp'});
    close CONFIG;

    ###  Set up ownerships.
    system "chown -R $webmaster $root/html/webmail        > /dev/null 2>&1";
    system "chgrp -R $group     $root/html/webmail        > /dev/null 2>&1";
    system "chgrp -R webd       $root/html/webmail/config > /dev/null 2>&1";
    system "chgrp -R webd       $root/html/webmail/data   > /dev/null 2>&1";

    ###  Set up sane permissions.
    system "chmod 770 $root/html/webmail/config           > /dev/null 2>&1";
    system "chmod 500 $root/html/webmail/config/conf.pl   > /dev/null 2>&1";
    system "chmod 330 $root/html/webmail/data             > /dev/null 2>&1";
    system "chmod 660 $root/html/webmail/data/*           > /dev/null 2>&1";


    #
    # Potentially add the hosts.allow entry for Squirrelmail.
    #
    require "../Security/Security.pl";

    @TMP = &GetACL('allow', 'simap');

    foreach my $t (@TMP) {
      goto here if ($t eq '127.0.0.1');
    }

    AddACL('127.0.0.1', 'Allow', 'simap');

here:

    #
    # Restart xinetd as xinetd.d/imap may have just been added.
    #
    system "/etc/init.d/xinetd restart > /dev/null 2>&1";
  }


  ### ##########################################################################
  ### Set final owernerships and permissions.
  ###

  $UID  =  getpwnam($webmaster);
  $GID  =  getgrnam($group);

  ###  Change the ownerships and set-group-id the directories.
  ###    - Only do the SSL stuff if we are doing an SSL site.
  chown $UID, $GID, "$root/cgi-bin", "$root/html", "$root/logs";
  chown $UID, $GID, "$root/ssl"  if ($port eq ":443");
  chmod 0700,            "$root/ssl"  if ($port eq ":443");
  chmod 0755,            "$root";
  chmod oct(2775),       "$root/cgi-bin", "$root/html", "$root/logs";


  ### ##########################################################################
  ### Finally, restart the daemon and go back to the index page.
  ### 
  %ENV_RESTORE  =  &clean_env();
  system("/etc/init.d/httpd restart > /dev/null 2>&1");
  &restore_env(%ENV_RESTORE);

  ###  I'm on my way... just set me free, home sweet home!
  &redirect("index.cgi");

